import sys
import math
import time
import PyQt5.QtWidgets as widgets
import PyQt5.QtCore as core
import PyQt5.QtGui as gui
import PyQt5.uic as uic
import types

colors = [core.Qt.black, core.Qt.red, core.Qt.green, core.Qt.blue, core.Qt.yellow, core.Qt.magenta, core.Qt.gray]
radius = 40.0
refreshRate = 50 # milliseconds

class Puck:
    def __init__(self, pos, touchPos, id):
        self.positions = [pos] # the past few positions
        self.times = [time.perf_counter()] # the times at which the past few events arrived
        self.idNumber = id
        self.velocityX = 0.0
        self.velocityY = 0.0
        self.touchPosition = touchPos # where the puck has been touched, relative to its center

pucks = []
    
def handleEvents(self, event):
    global lastEventTime
    if event.type() in [core.QEvent.TouchBegin, core.QEvent.TouchUpdate, core.QEvent.TouchEnd]:
        print(event.type())
        for touchPoint in event.touchPoints():
            print(touchPoint.state(), touchPoint.pos(), touchPoint.id())
            if touchPoint.state() == core.Qt.TouchPointPressed:
                newPuckToBeCreated = True
                for puck in pucks:
                    dx = touchPoint.pos().x() - puck.positions[0].x() # Here, puck.positions contains exactly one entry
                    dy = touchPoint.pos().y() - puck.positions[0].y()
                    distance = math.sqrt(dx**2 + dy**2)          
                    if distance < 2.0 * radius: # don't create new pucks too close to existing ones
                        newPuckToBeCreated = False
                    if distance < radius:
                        puck.touchPosition = core.QPointF(dx, dy)
                        puck.idNumber = touchPoint.id()
                        break
                if newPuckToBeCreated:
                    p = Puck(touchPoint.pos(), core.QPointF(), touchPoint.id())
                    pucks.append(p)
            elif touchPoint.state() == core.Qt.TouchPointMoved:
                for puck in pucks:
                    if puck.idNumber == touchPoint.id():
                        puck.times.append(time.perf_counter())
                        puck.positions.append(core.QPointF(touchPoint.pos().x() - puck.touchPosition.x(), touchPoint.pos().y() - puck.touchPosition.y()))
                        if len(puck.times) > 10: # do not store more than the past ten
                            del puck.positions[0]
                            del puck.times[0]
            elif touchPoint.state() == core.Qt.TouchPointReleased:
                for puck in pucks:
                    if puck.idNumber == touchPoint.id():
                        dt = puck.times[-1] - puck.times[0] + 0.001 # small offset to never divide by zero
                        puck.velocityX = (puck.positions[-1].x() - puck.positions[0].x()) / dt
                        puck.velocityY = (puck.positions[-1].y() - puck.positions[0].y()) / dt
                        puck.idNumber = -1                                                                          
                        if len(puck.positions) > 0: # only leave the latest timed position
                            del puck.positions[0:-1]
                            del puck.times[0:-1]
        self.update()
    elif event.type() == core.QEvent.Enter: # get rid of the regular mouse pointer
        self.setCursor(core.Qt.BlankCursor)
    elif event.type() == core.QEvent.Leave:
        self.setCursor(core.Qt.ArrowCursor)
    elif event.type() == core.QEvent.Paint:
        painter.begin(self)
        n = 0
        for puck in pucks:
            c = colors[n % len(colors)]
            n += 1
            painter.setBrush(c)
            painter.setPen(core.Qt.NoPen)
            painter.drawEllipse(puck.positions[-1], radius, radius)
        painter.end()
    return True

def animate():
    for puck in pucks:
        if puck.idNumber == -1: # is not under finger
            x = puck.positions[-1].x() + puck.velocityX * refreshRate / 1000.0
            y = puck.positions[-1].y() + puck.velocityY * refreshRate / 1000.0
            # bounce off borders
            if x < radius:
                x = radius
                puck.velocityX = -puck.velocityX
            elif x > w.widget.width() - radius:
                x = w.widget.width() - radius
                puck.velocityX = -puck.velocityX
            if y < radius:
                y = radius
                puck.velocityY = -puck.velocityY
            elif y > w.widget.height() - radius:
                y = w.widget.height() - radius
                puck.velocityY = -puck.velocityY
            puck.positions[-1] = core.QPointF(x, y)
    w.widget.update()

app = widgets.QApplication(sys.argv)
w = uic.loadUi("zehn_finger_hockey.ui")
painter = gui.QPainter()

timer = core.QTimer()
timer.timeout.connect(animate)
timer.start(refreshRate)

w.widget.setAttribute(core.Qt.WA_AcceptTouchEvents)
w.widget.event = types.MethodType(handleEvents, w.widget)
#w.setWindowState(core.Qt.WindowMaximized)
w.statusBar().hide()
w.show()
sys.exit(app.exec_())